Level = Core.class(Sprite)

function Level:init()
	application:setBackgroundColor(0x1234AA)

	-- set up a fullscreen 3D viewport
	local sw,sh=application:getContentWidth(),application:getContentHeight()
	self.view=D3.View.new(sw,sh,45,0.1,1000)
	self.scene=self.view:getScene() -- we will add all 3D models to it to build our 3D "scene"

	-- build a 100x100 floor plane (with normals)
	local gplane=D3.Mesh.new()
	gplane:setVertexArray{-100,0,-100, 100,0,-100, 100,0,100, -100,0,100}
	local tw,th=3200,3200
	gplane:setTextureCoordinateArray{0,0,tw,0,tw,th,0,th}
	gplane:setGenericArray(3,Shader.DFLOAT,3,4,{
		0,1,0,0,1,0,0,1,0,0,1,0,
	})
	gplane:setIndexArray{1,2,3,1,3,4}
	gplane:setTexture(Texture.new("textures/grass.png",true,{wrap=TextureBase.REPEAT}))
	gplane:updateMode(D3.Mesh.MODE_TEXTURE|D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW,0)

	-- some glb models
	local vegetation = {}
	local mushr01f = Glb.new("models/glbs", "shrooms.glb") -- model folder, model name
	local grass01f = Glb.new("models/glbs", "grass.glb")
	local pickvegetal, vegetal
	for i = 1, 64 do
		if i<=16 then pickvegetal=mushr01f vegetal="mush01"
		else pickvegetal=grass01f vegetal="grass"
		end
		vegetation[i] = G3DFormat.buildG3D(pickvegetal:getScene())
		vegetation[i].vegetal = vegetal
		vegetation[i]:updateMode(D3.Mesh.MODE_LIGHTING)
	end

	-- our 1st animated character (gdx/g3dj format with materials)
	self.m=buildGdx("models/characterLargeFemale/characterLargeFemale.json",
		{ modelpath="models/characterLargeFemale/", }
	)
	-- with two animations
	local animWalk=buildGdx("models/characterLargeFemale/walk.json",{})
	local animIdle=buildGdx("models/characterLargeFemale/idle.json",{})
	-- set default animation to idle
	-- note: in our files, the first animation in array is the T-pose, the second is the real anim
	D3Anim.setAnimation(self.m,animIdle.animations[2],"main",true)

	-- TIME TO SET UP OUR SCENE
	-- we scale our models
	local scale = 1.5
	for _, v in ipairs(vegetation) do -- glb
		if v.vegetal == "grass" then
			v:setScale(scale*math.random(1,4)*0.1,scale*math.random(2,7)*0.1,scale*math.random(1,4)*0.1)
		else -- mushrooms
			v:setScale(scale*math.random(5,8)*0.1,scale*math.random(5,9)*0.1,scale*math.random(5,8)*0.1)
		end
	end
	self.m:setScale(0.01, 0.01, 0.01) -- json (from fbx)
	-- we rotate some models
	for _, v in ipairs(vegetation) do v:setRotationY(math.random(0,360)) end
	-- we position our models
	for _, v in ipairs(vegetation) do
		if v.vegetal == "grass" then v:setPosition(math.random(-10,10), 0, math.random(-7,1))
		else v:setPosition(math.random(-15,15), 0, math.random(-15,1))
		end
	end
	self.m:setPosition(0, 0, 0)

	-- now we add the models to the scene
	self.scene:addChild(gplane)
	for _, v in ipairs(vegetation) do
		self.scene:addChild(v)
	end
	self.scene:addChild(self.m)
	-- and finally everything to stage
	self:addChild(self.view)

	-- light setup
	Lighting.setLight(5,10,10,0.3)
	Lighting.setLightTarget(0,2,0,30,45)

	-- INPUT: mouse down/up to start/stop walking
	self.walk=false
	self:addEventListener(Event.MOUSE_DOWN,function()
		--function D3Anim.setAnimation(model,anim,track,loop,transitionTime,speed)
		D3Anim.setAnimation(self.m,animWalk.animations[2],"main",true,0.5,1)
		self.walk=true
	end)
	self:addEventListener(Event.MOUSE_UP,function()
		D3Anim.setAnimation(self.m,animIdle.animations[2],"main",true,0.5,0.6)
		self.walk=false
	end)

	-- show a note on the top of screen to let the user know how to interact
	local text=TextField.new(nil,"Click to walk!")
	text:setTextColor(0xffff00)
	self:addChild(text) text:setPosition(10,20)

	local function BTNTXT(xposx, xposy, xwidth, xheight, xtext, xttf, xparent, xcatcher)
		local btn = ButtonTextP9UDDT.new({
			isautoscale=false,
			width=xwidth, height=xheight,
			pixelcolorup=0x7d0633, pixelcolordown=0x7ddd33, pixelalpha=0.5,
			text=xtext, ttf=xttf or gcf, textalpha=0.9,
			textcolorup=gtextcolor1, textcolordown=gtextcolor1outline,
			catcherx=xcatcher or 128,
		})
		btn:setPosition(xposx, xposy)
		xparent:addChild(btn)
		return btn
	end
	local btn01 = BTNTXT(64, 80, 100, 50, "MENU", nil, self, 8)
	btn01:addEventListener("clicked", function(e)
		scenemanager:changeScene("menu", 1, transitions[1], easings[1])
	end)
	self:addChild(btn01)

	self:addEventListener(Event.ENTER_FRAME, self.onEnterFrame, self)
	self:addEventListener("enterBegin", self.onTransitionInBegin, self)
	self:addEventListener("enterEnd", self.onTransitionInEnd, self)
	self:addEventListener("exitBegin", self.onTransitionOutBegin, self)
	self:addEventListener("exitEnd", self.onTransitionOutEnd, self)
end

-- THE GAME LOOP
local wang=0
function Level:onEnterFrame(e)
	-- Position our model according to its path angle
	if self.walk then wang+=2/3 end
	local an,r=^<wang,3
	local as,ac=math.sin(an)*r,math.cos(an)*r
	self.m:setRotationY(wang+90)
	self.m:setPosition(as,0.2,ac)
	-- Look at it
	self.view:lookAt(0,3,5,as,2,ac)
	-- Animation engine tick
	D3Anim.tick()
	-- Compute shadows
	Lighting.computeShadows(self.scene)
end

-- EVENT LISTENERS
function Level:onTransitionInBegin() end
function Level:onTransitionInEnd() end
function Level:onTransitionOutBegin()
	self:removeAllListeners()
--	D3Anim = nil
end
function Level:onTransitionOutEnd() end
